mod helper;
pub mod layout;

use cdtypes::types::{cdstring::DString, dir_record::DirectoryRecord};
use tool_helper::Error;

pub struct CDDesc {
    system_area: SystemArea,
    pvd:         PrimaryVolumeDescriptor,
    pub root:    Directory
}

impl CDDesc {
    pub fn new() -> CDDesc {
        match Directory::new("root") {
            Ok(root)   => CDDesc{system_area: SystemArea::new(), pvd: PrimaryVolumeDescriptor::new(), root},
            Err(error) => panic!("Creating root directory failed with: {}", error)
        }
    }

    pub fn get_memory_layout(&self) -> layout::MemoryLayout {
        layout::DefaultLayout::new(self)
    }

    pub fn get_memory_layout_mut(&mut self) -> layout::MemoryLayoutMut {
        layout::DefaultLayoutMut::new(self)
    }

    pub fn calculate_lbas(&mut self) {
        Self::for_each_dir_mut(&mut self.root, &|dir| {dir.update_content_size();});

        // Now layout iterate?
    }

    fn for_each_dir_mut<T: Fn(&mut Directory)>(dir: &mut Directory, function: &T) {
        function(dir);
        for sub_dir in &mut dir.dirs {
            Self::for_each_dir_mut(sub_dir, function);
        }
    }
}

pub struct SystemArea {
}

impl SystemArea {
    pub fn new() -> SystemArea {
        SystemArea{}
    }
}

pub struct PrimaryVolumeDescriptor {
}

impl PrimaryVolumeDescriptor {
    pub fn new() -> PrimaryVolumeDescriptor {
        PrimaryVolumeDescriptor{}
    }
}

pub struct Directory {
    pub name:       DirectoryName,
    pub properties: Properties,
        files:      Vec<File>,
        dirs:       Vec<Directory>
}

impl Directory {
    pub fn new(dir_name: &str) -> Result<Directory, Error> {
        Ok(Directory{name: DirectoryName::from_str(dir_name)?, properties: Properties::default(), files: Vec::new(), dirs: Vec::new()})
    }

    pub fn add_dir(&mut self, dir: Directory) {
        self.dirs.push(dir);
    }

    pub fn add_file(&mut self, file: File) {
        self.files.push(file);
    }

    fn update_content_size(&mut self) {
        let dir_member     = helper::collect_directory_record_member(&self.files, &self.dirs);
        let mut size_bytes = 0;

        for member in dir_member {
            size_bytes += DirectoryRecord::calculate_size_for(member.get_name().as_str(), true);
        }

        self.properties.overwrite_size_bytes = Some(size_bytes);
    }

    fn _update_content(&mut self) {
        let dir_member = helper::collect_directory_record_member(&self.files, &self.dirs);

        println!("{} updating content", self.name.as_string().unwrap_or("<No name>".to_owned()));
        for member in dir_member {
            println!(">>> {}", member.get_name());
        }        
    }
}

impl std::fmt::Display for Directory {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Directory: {}", self.name)
    }    
}

pub struct File {
    pub name:       FileName,
    pub properties: Properties,
        _content:    Vec<u8>
}

impl File {
    pub fn new(file_name: &str) -> Result<File, Error> {
        Ok(File{name: FileName::from_str(file_name)?, properties: Properties::default(), _content: Vec::new()})
    }
}

impl std::fmt::Display for File {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "File: {}", self.name)
    }
}

pub struct DirectoryName {
    name: DString<8>,
    len:  usize,
}

impl DirectoryName {
    pub fn from_str(dir_name: &str) -> Result<DirectoryName, Error> {
        let dir_name = dir_name.to_uppercase();
        Ok(DirectoryName{name: DString::from_str(dir_name.as_ref())?, len: dir_name.len()})
    }

    pub fn as_string(&self) -> Option<String> {
        Some(self.as_str()?.to_owned())
    }

    fn as_str(&self) -> Option<&str> {
        dstring_as_str(&self.name, self.len)
    }
}

impl std::fmt::Display for DirectoryName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.as_str().unwrap_or("<???>"))
    } 
}

pub struct FileName {
    name:    DString<8>,
    len:     usize,
    ext:     Option<DString<3>>,
    ext_len: Option<usize>
}

impl FileName {
    pub fn from_str(file_name: &str) -> Result<FileName, Error> {
        let file_name = file_name.to_uppercase();
        let (name, ext) = {
            let mut sub_str = file_name.split('.');
            (Error::ok_or_new(sub_str.next(), || "File name can't be emplty".to_owned())?, sub_str.next())
        };
        let (ext, ext_len) = {
            match ext {
                Some(ext) => (Some(DString::from_str(ext)?), Some(ext.len())),
                None      => (None, None)
            }   
        };

        Ok(FileName{name: DString::from_str(name)?, len: name.len(), ext, ext_len})
    }

    pub fn as_string(&self) -> Option<String> {
        if let Some(ext) = &self.ext {
            Some(format!("{}.{};1", dstring_as_str(&self.name, self.len)?, dstring_as_str(&ext, self.ext_len.unwrap_or(0))?))
        }

        else {
            Some(format!("{}", dstring_as_str(&self.name, self.len)?))
        }
    }
}

impl std::fmt::Display for FileName {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.as_string().unwrap_or("<???>".to_owned()))
    }
}

pub struct Properties {
    pub lba:                  usize,
    pub overwrite_size_bytes: Option<usize>,
    pub is_hidden:            bool
}

impl Default for Properties {
    fn default() -> Self {
        Properties{lba: 0, overwrite_size_bytes: None, is_hidden: false}
    }
}

fn dstring_as_str<const SIZE: usize>(string: &DString<SIZE>, len: usize) -> Option<&str> {
    match std::str::from_utf8(&string.as_raw()[0..len]) {
        Ok(str) => Some(str),
        Err(_)  => None,
    }
}